// TODO: send update progress events to the ui and create progress bar

// CONFIG

var default_config = {
  game_key: 'unknown',
  update_host: 'http://update.r2.redrobot.com',
  r2_dir_name: '.r2',
  current_dir_name: 'current',
  client_dir_name: 'client',
  manifest_file_name: 'manifest.js',
  update_interval: 60 * 60 * 24
};

// global context for now
var ctx = null;

// INIT

window.onload = function () {
  document.addEventListener('deviceready', main, false);
};

// MAIN

function main() {

  // store data so r2 client can retrieve when it is loaded
  sessionStorage.setItem("r2_launch_url", r2.launch_url);
  sessionStorage.setItem("r2_dev_token", r2.dev_token);
  sessionStorage.setItem("r2_game_token", r2.game_token);
  sessionStorage.setItem("r2_game_key", r2.game_key);
  sessionStorage.setItem("r2_player_token", r2.player_token);
  log('host config:', dump_obj(r2));

  // load and dump config to verify everything is setup correctly
  extend(default_config, config || {});
  config = default_config;
  extend(config, r2);
  log('config:', dump_obj(config));

  // setup ui
  setup_ui();

  // start main flow
  flow_load();
  //show_status("Loading Content");
  //show_splash("Loading");
  //show_error({msg:"none"}, "Failed to update");
}

// FLOW

function flow_load() {
  // init global ctx for each time through the flow so restarts are not memoized
  ctx = {};

  show_splash('Loading');
  is_online(function() {
    flow_should_check_for_update(function() {
      show_status('Checking for update');
      flow_has_latest_version(function () {
        launch_client();
      }, function (error) {
        flow_update_client(function (entry) {
          launch_client();
        }, function (error) {
          show_error(error, 'Unable to update')
        });
      });
    }, function() {
      launch_client();
    });
  }, function(error) {
    show_error(error, 'Not online')
  });
}

function flow_should_check_for_update(success_cb, error_cb) {
  var updated_at = get_updated_at();
  log('updated_at', updated_at);
  var now = new Date().getTime();
  var update_interval = config.update_interval;
  if (updated_at && (updated_at > (now - update_interval)))  {
    error_cb();
  } else {
    success_cb();
  }
}

function flow_has_latest_version(success_cb, error_cb) {
  load_js_file(':current_dir/:manifest_file', function (current_obj) {
    var current_manifest = current_obj.manifest;
    ctx.manifest = current_manifest;
    ctx.current_manifest = current_manifest;
    log('current version', current_manifest.version);
    resolve_path(':client_dir/:random_dir', function (target_path) {
      download_and_load_js_file(':update_host/:game_key/:manifest_file', target_path, function (remote_obj) {
        var remote_manifest = remote_obj.manifest;
        ctx.remote_manifest = remote_manifest;
        log('remote version', remote_manifest.version);
        if (remote_manifest.force_update === true || remote_manifest.version !== current_manifest.version) {
          error_cb({msg: 'needs update'});
        } else {
          success_cb();
        }
      }, error_cb);
    }, error_cb);
  }, error_cb);
}

function flow_update_client(success_cb, error_cb) {
  show_status('updating');
  resolve_path(':client_dir/:random_dir', function (target_path) {
    download_and_load_js_file(':update_host/:game_key/:manifest_file', target_path, function (obj) {
      var remote_manifest = obj.manifest;
      ctx.manifest = remote_manifest;
      ctx.remote_manifest = remote_manifest;
      log('remote version', remote_manifest.version);
      var total = remote_manifest.files.length;
      var current = 0;
      show_status('updating - ' + current + '/' + total);
      each(remote_manifest.files, function (idx, file_name, next) {
        download_file(build_path(':update_host/:game_key', file_name), build_path(target_path, base_path_segment(file_name)), function (entry) {
          current = current + 1;
          show_status('updating - ' + current + '/' + total);
          next();
        }, error_cb);
      }, function () {
        delete_dir(':client_dir/:current_dir', function () {
          move_dir(target_path, ':client_dir/:current_dir', function (entry) {
            set_updated_at();
            success_cb(entry);
          }, error_cb);
        }, error_cb);
      });
    }, error_cb);
  }, error_cb);
}

function launch_client() {
  show_status('launching');
  load_js_file(':current_dir/:manifest_file', function (current_obj) {
    var current_manifest = current_obj.manifest;
    ctx.manifest = current_manifest;
    ctx.current_manifest = current_manifest;
    show_status('launching - v' + ctx.current_manifest.version);
    verify_path(build_path(':current_dir', ctx.manifest.main_file), function (client_path) {
      resolve_path(':current_dir', function(current_path) {
        sessionStorage.setItem("r2_base_url", current_path);
        log('launching', client_path);
        navigator.app.loadUrl(client_path, {'openexternal': false, 'clearhistory': true});
      }, function(error) {
        show_error(error, 'unable to launch');
      })
    }, function(error) {
      show_error(error, 'unable to launch');
    });
  }, function(error) {
    show_error(error, 'unable to load');
  });
}

function flow_retry() {
  delete_dir(':client_dir', function () {
    flow_load();
  }, function(error) {
    show_error(error, 'unable to update');
  });
}

// VIEW

function show_splash(msg) {
  show_view('splash');
  var element = document.getElementById('splash_msg');
  element.innerHTML = msg;
}

function show_status(msg) {
  log(msg);
  show_view('status');
  var element = document.getElementById('status_msg');
  element.innerHTML = msg;
}

function show_error(error, msg) {
  log('error:', error.msg, ' - ', msg);
  show_view('error');
  var element_msg = document.getElementById('error_msg');
  element_msg.innerHTML = msg;

  var element_button_msg = document.getElementById('button_msg');
  element_button_msg.innerHTML = "Press to Retry";
}

// FUNC

function is_online(success_cb, error_cb) {
  if(Connection.NONE === navigator.network.connection.type) {
    error_cb({msg: 'no connection'});
  } else {
    success_cb();
  }
}

// might want to strip all schemes
function strip_protocol(filePath) {
  var proto = 'file://';
  if (filePath.substring(0, proto.length) === proto) {
    return filePath.substring(proto.length);
  } else {
    return filePath;
  }
}

function build_path() {
  var args = Array.prototype.slice.call(arguments);
  var segments = [];
  var arg = null;
  for (var i = 0; i < args.length; i++) {
    arg = args[i];
    if (arg.hasOwnProperty('fullPath')) {
      segments.push(arg.fullPath);
    } else {
      segments.push(arg);
    }
  }
  return segments.join('/');
}

function read_file(file_path, success_cb, error_cb) {
  get_file(file_path, function (entry) {
    entry.file(function (file) {
      var reader = new FileReader();
      reader.onloadend = function (evt) {
        success_cb(evt.target.result);
      };
      reader.readAsText(file);
    }, function(error) {
      error_cb({msg: 'unable to get file ' + entry.fullPath});
    });
  }, error_cb);
}

function load_js_file(file_path, success_cb, error_cb) {
  read_file(file_path, function (data) {
    try {
      var func = new Function('exports', data);
      var exports = {};
      func(exports);
      success_cb(exports);
    } catch (e) {
      error_cb({msg: 'error loading js file ' + e});
    }
  }, error_cb);
}

function download_and_load_js_file(source, target_dir, success_cb, error_cb) {
  download_file(source, target_dir, function (entry) {
    load_js_file(entry.fullPath, function (obj) {
      success_cb(obj);
    }, error_cb);
  }, error_cb);
}

function get_file(file_path, success_cb, error_cb) {
  resolve_path(file_path, function (fullpath) {
    window.resolveLocalFileSystemURI(fullpath, success_cb, function(error) {
      error_cb({msg: 'unable to get file ' + fullpath})
    });
  }, error_cb);
}

function get_dir(dir_path, success_cb, error_cb) {
  get_file(dir_path, success_cb, error_cb);
}

function download_file(source, target_dir, success_cb, error_cb) {
  var target = null;
  assure_dir(target_dir, function (entity) {
    target = strip_protocol(build_path(entity, last_path_segment(source)));
    resolve_path(source, function (source_path) {
      resolve_path(target, function (target_path) {
        new FileTransfer().download(source_path, target_path, function (entry) {
          success_cb(entry);
        }, function(error) {
          error_cb({msg: 'unable to download file ' + source_path});
        });
      }, error_cb);
    }, error_cb);
  }, error_cb);
}

function delete_dir(path, success_cb, error_cb) {
  resolve_file(path, function (entry) {
    entry.removeRecursively(function (entry) {
      success_cb();
    }, function (error) {
      error_cb({msg: 'unable to remove path ' + entry.fullPath});
    });
  }, function (error) {
    // if the path can't be resolved then it doesn't need to be deleted
    success_cb();
  });
}

function move_dir(source_path, target_path, success_cb, error_cb) {
  resolve_dir(source_path, function (source_dir) {
    resolve_path(target_path, function (target_fullpath) {
      var segments = split_path(target_fullpath);
      resolve_dir(segments[0], function (parent_dir) {
        source_dir.moveTo(parent_dir, segments[1], function (entry) {
          success_cb(entry);
        }, function (error) {
          error_cb({msg: 'unable to move directory'});
        });
      }, error_cb);
    }, error_cb);
  }, error_cb);
}

function get_updated_at() {
  return localStorage.getItem("updated_at");
}

function set_updated_at() {
  localStorage.setItem("updated_at", new Date().getTime());
}

// PATH

function get_root_dir(success_cb, error_cb) {
  window.requestFileSystem(LocalFileSystem.PERSISTENT, 0, function (filesystem) {
    success_cb(filesystem.root);
  }, function (error) {
    error_cb({msg: 'unable to get root directory'});
  });
}

function get_root_dir_path(success_cb, error_cb) {
  get_root_dir(function (entry) {
    success_cb(entry.fullPath);
  }, error_cb);
}

function get_current_dir(success_cb, error_cb) {
  get_client_dir(function (entry) {
    get_current_dir_name(function (name) {
      entry.getDirectory(name, {create:true}, function (entry) {
        success_cb(entry);
      }, function(error) {
        error_cb({msg: 'could not get directory ' + entry.fullPath + '/' + name});
      });
    }, error_cb);
  }, error_cb);
}

function get_current_dir_path(success_cb, error_cb) {
  get_current_dir(function (entry) {
    success_cb(entry.fullPath);
  }, error_cb);
}

function get_current_dir_name(success_cb, error_cb) {
  success_cb(config.current_dir_name);
}

function get_r2_dir(success_cb, error_cb) {
  get_root_dir(function (entry) {
    get_r2_dir_name(function (name) {
      entry.getDirectory(name, {create:true}, function (entry) {
        success_cb(entry);
      }, function(error) {
        error_cb({msg: 'could not get directory ' + entry.fullPath + '/' + name});
      });
    }, error_cb);
  }, error_cb);
}

function get_r2_dir_path(success_cb, error_cb) {
  get_r2_dir(function (entry) {
    success_cb(entry.fullPath);
  }, error_cb);
}

function get_r2_dir_name(success_cb, error_cb) {
  success_cb(config.r2_dir_name);
}

function get_client_dir(success_cb, error_cb) {
  get_r2_dir(function (entry) {
    get_client_dir_name(function (name) {
      entry.getDirectory(name, {create:true}, function (entry) {
        success_cb(entry);
      }, function(error) {
        error_cb({msg: 'could not get directory ' + entry.fullPath + '/' + name});
      });
    }, error_cb);
  }, error_cb);
}

function get_client_dir_path(success_cb, error_cb) {
  get_client_dir(function (entry) {
    success_cb(entry.fullPath);
  }, error_cb);
}

function get_client_dir_name(success_cb, error_cb) {
  success_cb(config.client_dir_name);
}

function get_manifest_file_name(success_cb, error_cb) {
  success_cb(config.manifest_file_name);
}

function get_random_dir_name(success_cb, error_cb) {
  success_cb(guid());
}

function get_update_host_path(success_cb, error_cb) {
  success_cb(config.update_host);
}

function get_game_key_name(success_cb, error_cb) {
  success_cb(config.game_key);
}

function verify_path(path, success_cb, error_cb) {
  resolve_path(path, function (path) {
    window.resolveLocalFileSystemURI(path, function (entry) {
      success_cb(path);
    }, function (error) {
      error_cb({msg: 'path does not exist'});
    });
  }, error_cb);
}

function assure_dir(dir_path, success_cb, error_cb) {
  resolve_path(dir_path, function (path) {
    window.resolveLocalFileSystemURI(dir_path, function (entry) {
      success_cb(entry);
    }, function (error) {
      var segments = split_path(dir_path);
      assure_dir(segments[0], function (entry) {
        entry.getDirectory(segments[1], {create:true}, function (entry) {
          success_cb(entry);
        }, function(error) {
          error_cb({msg: 'unable to get directory'})
        });
      }, error_cb);
    });
  }, error_cb);
}

function resolve_file(path, success_cb, error_cb) {
  resolve_path(path, function (full_path) {
    get_file(full_path, success_cb, error_cb);
  }, error_cb);
}

function resolve_dir(path, success_cb, error_cb) {
  resolve_file(path, success_cb, error_cb);
}

function resolve_path(path, success_cb, error_cb) {
  if (path == null || path == '') {
    error_cb({msg: "path cannot be empty"});
  }

  var result_segments = [];
  var func_name = null;
  var func = null;

  each(path.split('/'), function (idx, segment, next_cb) {
    if (segment.substring(0, 1) === ':') {
      func_name = segment.substring(1);
      if (result_segments.length === 0) {
        func = window['get_' + func_name + '_path'];
      } else {
        func = window['get_' + func_name + '_name'];
      }
      if (func && typeof func === 'function') {
        func(function (value) {
          result_segments.push(value);
          next_cb();
        }, function (error) {
          error_cb({msg: 'error calling lookup function ' + func_name});
        });
      } else {
        error_cb({msg: 'unable to resolve ' + func_name});
      }
    } else {
      result_segments.push(segment);
      next_cb();
    }
  }, function () {
    success_cb(result_segments.join('/'));
  });
}

function last_path_segment(path) {
  return split_path(path)[1];
}

function base_path_segment(path) {
  return split_path(path)[0];
}

function split_path(path) {
  var idx = path.lastIndexOf('/');
  if (idx) {
    return [path.substring(0, idx), path.substring(idx + 1)];
  } else {
    return [null, path];
  }
}

// MISC

memoize('read_file');
memoize('load_js_file');
memoize('get_file');
memoize('get_root_dir');
memoize('get_r2_dir');
memoize('get_random_dir_name');
memoize('assure_dir');
memoize('resolve_path');

function memoize(func_name) {
  var func = window[func_name];
  window[func_name] = function() {
    var args = Array.prototype.slice.call(arguments);
    var arg = null;
    var success_cb = null;
    var error_cb = null;

    // func(arg, success_cb(success), error_cb(error))
    if (args.length === 3) {
      arg = args[0];
      success_cb = args[1];
      error_cb = args[2];
    }
    // func(success_cb(success), error_cb(error))
    else {
      success_cb = args[0];
      error_cb = args[1];
    }

    var key = func.name + "_" + arg;

    if (key in ctx) {
      success_cb(ctx[key]);
    } else {
      if (arg) {
        func(arg, function(result) {
          ctx[key] = result;
          success_cb(result);
        }, error_cb);
      } else {
        func(function(result) {
          ctx[key] = result;
          success_cb(result);
        }, error_cb);
      }
    }
  };
}

function each(collection, next_cb, done_cb) {
  var len = collection.length;
  (function loop_at(i) {
    if (i < len) {
      next_cb(i, collection[i], function () {
        loop_at(i + 1);
      });
    } else {
      done_cb();
    }
  })(0);
}

function dump_obj(obj) {
  var props = [];
  for (prop in obj) {
    props.push(prop + ": " + obj[prop]);
  }
  return '{' + props.join(', ') + '}';
}

function log() {
  console.log('[spark] ' + Array.prototype.slice.call(arguments).join(' '));
}

function show_view(id) {
  var elements = document.getElementsByClassName('screen');
  for (var i = 0; i < elements.length; i++) {
    var element = elements[i];
    remove_class(element, 'active_screen');
  }

  var element = document.getElementById(id);
  add_class(element, 'active_screen');
}

function setup_ui() {
  var retry_button = document.getElementById('retry_button');
  retry_button.onclick = function() {
    flow_retry();
  };
}

function add_class(elm, class_name) {
  //elm.classList.add(class_name);
  var className = elm.className;
  if (className.indexOf(class_name) === -1) {
    elm.className += ' ' + class_name;
  }
}

function remove_class(elm, class_name) {
  //var className = elm.className;
  elm.className = elm.className.replace(class_name, '').trim();
  //elm.classList.remove(class_name);
}

// http://andrewdupont.net/2009/08/28/deep-extending-objects-in-javascript/
function extend(destination, source) {
  for (var property in source) {
    if (source[property] && source[property].constructor && source[property].constructor === Object) {
      destination[property] = destination[property] || {};
      arguments.callee(destination[property], source[property]);
    } else {
      destination[property] = source[property];
    }
  }
  return destination;
}

// http://note19.com/2007/05/27/javascript-guid-generator/
function S4() {
  return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
}
function guid() {
  return (S4() + S4() + "-" + S4() + "-" + S4() + "-" + S4() + "-" + S4() + S4() + S4());
}
